"""
Command Line Interface for sequence comparison API
"""
from typing import Tuple

from click import group, argument, option, Choice, get_current_context

from pyaliner.align import CLASSIC, COMPACT
from pyaliner.compare import PAIRED, INLINED, compare_in_true_pred_from_file, compare_true_pred_from_file, \
    compare_true_pred_from_files, compare_in_true_pred_from_files

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
ALI_HELP = 'what alignment algorithm to use:\ntoken to token (=classic) or\nmulti-token to multi-token (=compact)'
VIEW_HELP = 'what view to show:' \
            '\nfull line against full line (=paired) or\nsingle line with embedded differences (=inline)'


@group()
def pyaliner(context_settings=CONTEXT_SETTINGS):
    pass


@pyaliner.command(help='compares true and predicted sequences from either 1 paired file or 2 single files')
@argument('paths', nargs=-1)
@option('--alignment', default=COMPACT, show_default=True, type=Choice([CLASSIC, COMPACT]), help=ALI_HELP)
@option('--view', default=PAIRED, show_default=True, type=Choice([PAIRED, INLINED]), help=VIEW_HELP)
def tp(paths: Tuple[str, ...], alignment: str, view: str):
    if len(paths) == 1:

        compare_true_pred_from_file(paths[0], alignment, view)

    elif len(paths) == 2:

        compare_true_pred_from_files(*paths, alignment, view)

    else:

        get_current_context() \
            .fail(f'This commands accepts either a single file or two files but [{len(paths)}] paths were given')


@pyaliner.command(help='compares input, true and predicted sequences from either 1 triplet file or 3 single files')
@argument('paths', nargs=-1)
@option('--alignment', default=COMPACT, show_default=True, type=Choice([CLASSIC, COMPACT]), help=ALI_HELP)
def itp(paths: Tuple[str, ...], alignment: str):
    if len(paths) == 1:

        compare_in_true_pred_from_file(paths[0], alignment)

    elif len(paths) == 3:

        compare_in_true_pred_from_files(*paths, alignment)

    else:

        get_current_context() \
            .fail(f'This commands accepts either a single file or three files but [{len(paths)}] were given')


if __name__ == '__main__':
    pyaliner()



"""
Functions to display visually rich comparisons between sequential data
"""
import os
from itertools import zip_longest
from shutil import get_terminal_size
from typing import Tuple, Iterable, Sequence

from pypey import pype
from rich.console import Console

from pyaliner.align import align, COMPACT, JOIN, GAP, Seq

FILL = '∷'
ELLIPSIS = '[orange3]...[/orange3]'
WHITE = 'grey93'
BOLD_BLUE = 'b #0EADF1'
BOLD_RED = 'b #F1520E'
DIM_MAGENTA = 'dim bright_magenta'
MAGENTA = 'bright_magenta'
GREY = 'grey70'

SCREEN_WIDTH = 100
SPACE_N = 1
ELLIPSIS_N = 3
MARGIN_N = 3

DEFAULT_CONSOLE = Console(highlight=False)


def in_terminal(comparisons: Iterable[str], console: Console = DEFAULT_CONSOLE):
    """
    Displays comparisons with a pager in the terminal

    :param comparisons: collection of comparison lines
    :param console: console to print to terminal with
    :return: nothing, but displays comparison in terminal
    """

    # return tuple(comparisons)
    if not os.environ.get('PAGER'):
        os.environ['PAGER'] = 'less -r'

    with console.pager(styles=True):
        for line in comparisons:
            console.print(line)


def rich_paired_in_true_pred(alignments: Iterable[Tuple[Seq, Seq, Seq]]) -> Iterable[str]:
    """
    Creates rich visual comparison strings for a collection of triples of input, true and predicted sequences, aligned
    against each other

    :param alignments: Triples of tuples of strings representing inputs, ground-truths and predictions
    :return: a collection of rich strings for terminal display
    """
    screen_width, _ = get_terminal_size()

    return (pype(alignments)
            .map(lambda in_seq, true, pred: _single_in_true_pred_paired(in_seq, true, pred, screen_width))
            .map(''.join))


def rich_paired_true_pred(alignments: Iterable[Tuple[Seq, Seq]]) -> Iterable[str]:
    """
    Creates rich visual comparison strings for a collection of pairs of true and predicted sequences, aligned
    against each other

    :param alignments: Pairs of tuples of strings representing ground-truths and predictions
    :return: a collection of rich strings for terminal display
    """
    screen_width, _ = get_terminal_size()

    return pype(alignments).map(lambda true, pred: _single_true_pred_paired(true, pred, screen_width)).map(''.join)


def rich_inlined_true_pred(alignments: Iterable[Tuple[Seq, Seq]]) -> Iterable[str]:
    """
    Creates rich visual comparison strings for a collection of pairs of true and predicted sequences, embedded in the
    same line

    :param alignments: Pairs of tuples of strings representing ground-truths and predictions
    :return: a collection of rich strings for terminal display
    """
    for ref, rec in alignments:
        ui = []

        for ref_token, rec_token in zip(ref, rec):

            if ref_token == rec_token:

                ui.append(_in_white(ref_token))

            elif rec_token == GAP:

                ui.append(f'{_in_dim_magenta("[")}{_in_strike_red(ref_token)}{_in_dim_magenta("]")}')

            elif ref_token == GAP:

                ui.append(f'{_in_dim_magenta("[")}{_in_magenta("↗ ")}{_in_red(rec_token)}{_in_dim_magenta("]")}')

            else:

                ui.append(f'{_in_dim_magenta("[")}{_in_blue(ref_token)} {_in_red(rec_token)}{_in_dim_magenta("]")}')

        yield f' {" ".join(ui)}\n'.replace(JOIN, _in_white(JOIN))


def _single_in_true_pred_paired(in_seq: Sequence[str], true: Sequence[str], pred: Sequence[str], screen_width: int):
    in_ui, true_ui, pred_ui = [], [], []

    length = 0

    for in_token, true_token, pred_token in zip_longest(in_seq, true, pred, fillvalue=FILL):

        true_token, pred_token = _realign_true_pred(true_token, pred_token)

        pad = max(len(in_token), len(JOIN.join(true_token)), len(JOIN.join(pred_token)))

        if (length + pad + SPACE_N + ELLIPSIS_N + SPACE_N + MARGIN_N) >= screen_width:
            in_ui.append(f'{ELLIPSIS}')
            true_ui.append(f'{ELLIPSIS}')
            pred_ui.append(f'{ELLIPSIS}')

            yield f" {' '.join(in_ui)}\n {' '.join(true_ui)}\n {' '.join(pred_ui)}\n\n".replace(JOIN, _in_grey(JOIN))

            in_ui = [ELLIPSIS]
            true_ui = [ELLIPSIS]
            pred_ui = [ELLIPSIS]

            length = ELLIPSIS_N + SPACE_N

        in_colour_fn = _in_white if in_token == JOIN.join(true_token) and in_token == JOIN.join(
            pred_token) else _in_blue

        coloured_true_token, coloured_pred_token = _recolour(in_token, true_token, pred_token)

        in_ui.append(in_colour_fn(f'{in_token:^{pad}}'))
        true_ui.append(f'{coloured_true_token + " " * (max(0, len(in_token) - len(JOIN.join(true_token))))}')
        pred_ui.append(f'{coloured_pred_token + " " * (max(0, len(in_token) - len(JOIN.join(true_token))))}')

        length += pad + SPACE_N

    yield f" {' '.join(in_ui)}\n {' '.join(true_ui)}\n {' '.join(pred_ui)}\n".replace(JOIN, _in_grey(JOIN))


def _single_true_pred_paired(true: Sequence[str], pred: Sequence[str], screen_width: int) -> Iterable[str]:
    true_ui, pred_ui = [], []

    length = 0

    for true_token, pred_token in zip_longest(true, pred):

        pad = max(len(true_token), len(pred_token))

        if (length + pad + SPACE_N + ELLIPSIS_N + SPACE_N + MARGIN_N) >= screen_width:
            true_ui.append(f'{ELLIPSIS}')
            pred_ui.append(f'{ELLIPSIS}')

            yield f" {' '.join(true_ui)}\n {' '.join(pred_ui)}\n\n".replace(JOIN, _in_grey(JOIN))

            true_ui = [ELLIPSIS]
            pred_ui = [ELLIPSIS]

            length = ELLIPSIS_N + SPACE_N

        true_colour_fn = _in_white if true_token == pred_token else _in_blue
        pred_colour_fn = _in_white if true_token == pred_token else _in_red

        true_ui.append(true_colour_fn(f'{true_token:^{pad}}'))
        pred_ui.append(pred_colour_fn(f'{pred_token:^{pad}}'))

        length += pad + SPACE_N

    yield f" {' '.join(true_ui)}\n {' '.join(pred_ui)}\n".replace(JOIN, _in_grey(JOIN))


def _realign_true_pred(true: str, pred: str) -> Tuple[Seq, Seq]:
    unjoined_true, unjoined_pred = tuple(true.split(JOIN)), tuple(pred.split(JOIN))

    new_true, new_pred = [], []

    for true_bit, pred_bit in zip(*align(unjoined_true, unjoined_pred, kind=COMPACT)):
        pad = max(len(true_bit), len(pred_bit))

        new_true.append(f'{true_bit:^{pad}}')
        new_pred.append(f'{pred_bit:^{pad}}')

    return tuple(new_true[::-1]), tuple(new_pred[::-1])


def _recolour(in_seq: str, true: Seq, pred: Seq) -> Tuple[str, str]:
    if JOIN.join(true) == in_seq and JOIN.join(pred) == in_seq:
        return _in_white(' '.join(true)), _in_white(' '.join(pred))

    colored_true, coloured_pred = [], []

    for true_token, pred_token in zip(true, pred):
        colored_true.append(_in_blue(true_token))
        coloured_pred.append(_in_blue(pred_token) if true_token == pred_token else _in_red(pred_token))

    return ' '.join(colored_true[::-1]), ' '.join(coloured_pred[::-1])


def _in_white(token: str) -> str:
    return f'[{WHITE}]{token}[/{WHITE}]'


def _in_grey(token: str) -> str:
    return f'[{GREY}]{token}[/{GREY}]'


def _in_blue(token: str) -> str:
    return f'[{BOLD_BLUE}]{token}[/{BOLD_BLUE}]'


def _in_red(token: str) -> str:
    return f'[{BOLD_RED}]{token}[/{BOLD_RED}]'


def _in_dim_magenta(token: str) -> str:
    return f'[{DIM_MAGENTA}]{token}[/{DIM_MAGENTA}]'


def _in_magenta(token: str) -> str:
    return f'[{MAGENTA}]{token}[/{MAGENTA}]'


def _in_strike_red(token: str) -> str:
    return f'[s {BOLD_RED}]{token}[/s {BOLD_RED}]'


def _in_strong_red(token: str) -> str:
    return f'[u {BOLD_RED}]{token}[/u {BOLD_RED}]'



"""
Functions to align sequential data
"""

import re
from collections import deque
from typing import Tuple, Sequence, List

from edlib import align as edalign
from pypey import pype, Fn

Seq = Tuple[str, ...]

CLASSIC = 'classic'
COMPACT = 'compact'
NGRAM = 'ngram'

GAP = '⎵'
JOIN = '∙'

LEV = lambda left, right: float(left != right)

CIGAR_RE = re.compile(r'([=DIX])')

DELTA = 1e-6


def align(left: Sequence[str], right: Sequence[str], cost_fn: Fn[[str, str], float] = LEV, kind: str = CLASSIC) \
        -> Tuple[Seq, Seq]:
    """
    Aligns left and right sequences using the edit distance computation. Passing in a cost function allows
    varying the substitution, deletion and insertion costs. The type of alignment can be varied with the `kind`
    parameter, which can be `CLASSIC` OR `COMPACT`. `COMPACT` alignment merges adjacent mismatches into a single one,
    joined up with the `JOIN` symbol, which can show differences more clearly.

    In the current implementation, when a cost function other than `LEV` is passed in, or when the number of unique
    tokens in the left and right sequences is greater than 256, alignments slows down.
    
    :param left: left sequence of strings
    :param right: right sequence of strings
    :param cost_fn: function taking a left and a right token and returning a cost
    :param kind: the kind of of alignment, default is `CLASSIC`
    :return: a `tuple` of `str`s  with the left and right alignments
    """
    if cost_fn == LEV:
        return _fast_align(left, right) if kind == CLASSIC else _compact_align(left, right)

    return _flexible_align(left, right, cost_fn)[:-1]


def as_cigar(left: Sequence[str], right: Sequence[str]) -> str:
    """
    Returns a CIGAR-encoded version of the given alignment

    :param left: right alignment
    :param right: left alignment
    :return: CIGAR-encoded alignment
    """
    cigar = ''

    for left_token, right_token in zip(left, right):
        if left_token == right_token:

            if cigar and cigar[-1] == '=':
                *head, num, op = CIGAR_RE.split(cigar)[:-1]
                cigar = f'{"".join(head)}{int(num) + 1}{op}'
            else:
                cigar += '1='

        elif left_token == GAP:

            if cigar and cigar[-1] == 'D':
                *head, num, op = CIGAR_RE.split(cigar)[:-1]
                cigar = f'{"".join(head)}{int(num) + 1}{op}'
            else:
                cigar += '1D'

        elif right_token == GAP:

            if cigar and cigar[-1] == 'I':
                *head, num, op = CIGAR_RE.split(cigar)[:-1]
                cigar = f'{"".join(head)}{int(num) + 1}{op}'
            else:
                cigar += '1I'
        else:

            if cigar and cigar[-1] == 'X':
                *head, num, op = CIGAR_RE.split(cigar)[:-1]
                cigar = f'{"".join(head)}{int(num) + 1}{op}'
            else:
                cigar += '1X'

    return cigar


def compact(left_ali: Sequence[str], right_ali: Sequence[str]) -> Tuple[Seq, Seq]:
    """
    Takes a classic alignment and turns it into a compact one

    :param left_ali: left classic alignment
    :param right_ali: right classic alignment
    :return: compact alignment
    """
    new_left_ali, new_right_ali = [], []

    for left_token, right_token in zip(left_ali, right_ali):

        if (not new_left_ali and not new_right_ali) or left_token == right_token:
            new_left_ali.append(left_token)
            new_right_ali.append(right_token)

            continue

        last_left, last_right = new_left_ali[-1], new_right_ali[-1]

        if right_token == GAP:

            if last_left == GAP:

                new_left_ali.pop()
                new_left_ali.append(left_token)

            elif last_left != last_right:

                new_left_ali.append(_join(new_left_ali.pop(), left_token))

            else:

                new_left_ali.append(left_token)
                new_right_ali.append(right_token)

        elif left_token == GAP:

            if last_right == GAP:

                new_right_ali.pop()
                new_right_ali.append(right_token)

            elif last_left != last_right:

                new_right_ali.append(_join(new_right_ali.pop(), right_token))

            else:

                new_left_ali.append(left_token)
                new_right_ali.append(right_token)
        else:

            if last_left == GAP:

                new_left_ali.pop()
                new_left_ali.append(left_token)
                new_right_ali.append(_join(new_right_ali.pop(), right_token))

            elif last_right == GAP:

                new_right_ali.pop()
                new_right_ali.append(right_token)

                new_left_ali.append(_join(new_left_ali.pop(), left_token))

            else:

                new_left_ali.append(left_token)
                new_right_ali.append(right_token)

    return tuple(new_left_ali), tuple(new_right_ali)


def _fast_align(left: Sequence[str], right: Sequence[str]) -> Tuple[Seq, Seq]:
    """
    A fast implementation of the classic edit distances alignment, using `edlib`

    :param left: left sequence of strings
    :param right:  right sequence of strings
    :return: a `tuple` of `str`s  with the left and right alignments
    """
    cigar = None
    try:

        cigar = edalign(left, right, task='path')['cigar']

    except ValueError:

        cigar = as_cigar(*_flexible_align(left, right, LEV)[:-1])

    cigar = pype(CIGAR_RE.split(cigar)[:-1]).chunk(2)

    left_start, right_start = 0, 0
    left_ali, right_ali = [], []

    for num, op in cigar:
        num = int(num)

        if op == 'X' or op == '=':

            left_ali.extend(left[left_start: left_start + num])
            right_ali.extend(right[right_start: right_start + num])

            left_start, right_start = left_start + num, right_start + num

        elif op == 'I':

            left_ali.extend(left[left_start: left_start + num])
            right_ali.extend([GAP] * num)

            left_start += num

        else:
            left_ali.extend([GAP] * num)
            right_ali.extend(right[right_start: right_start + num])

            right_start += num

    return tuple(left_ali), tuple(right_ali)


def _compact_align(left: Sequence[str], right: Sequence[str]) -> Tuple[Seq, Seq]:
    """
    Aligns sequences by finding the classic edit distance alignment and then merging adjacent substitutions, insertions
    and deletions, joined up with the `JOIN` symbol

    :param left: left sequence of strings
    :param right:  right sequence of strings
    :return: a `tuple` of `str`s  with the left and right alignments
    """
    cigar = None

    try:

        cigar = edalign(left, right, task='path')['cigar']

    except ValueError:

        cigar = as_cigar(*_flexible_align(left, right, LEV)[:-1])

    cigar = pype(CIGAR_RE.split(cigar)[:-1]).chunk(2).eager()

    left_start, right_start = 0, 0
    left_ali, right_ali = [], []

    for num, op in cigar:
        num = int(num)

        if op == '=':

            left_ali.extend(left[left_start: left_start + num])
            right_ali.extend(right[right_start: right_start + num])

            left_start, right_start = left_start + num, right_start + num

        elif op == 'X':

            left_ali.append(JOIN.join(left[left_start: left_start + num]))
            right_ali.append(JOIN.join(right[right_start: right_start + num]))

            left_start, right_start = left_start + num, right_start + num

        elif op == 'I':

            left_ali.extend(left[left_start: left_start + num])
            right_ali.extend([GAP] * num)

            left_start += num

        else:
            left_ali.extend([GAP] * num)
            right_ali.extend(right[right_start: right_start + num])

            right_start += num

    return compact(left_ali, right_ali)


def _flexible_align(left: Sequence[str], right: Sequence[str], cost_fn: Fn[[str, str], float]) \
        -> Tuple[Seq, Seq, float]:
    """
    Slow version of the classic edit distance alignment supporting parametrisation of the cost matrix

    :param left: left sequence of strings
    :param right:  right sequence of strings
    :param cost_fn: function taking a left and a right token and returning a cost
    :return: a `tuple` of `str`s  with the left and right alignments
    """
    if not left or not right:
        raise Exception(f'{left} or {right} or both are empty')

    return _grid_align(left, right, cost_fn, _grid(left, right, cost_fn))


def _grid_align(left: Sequence[str], right: Sequence[str], cost_of: Fn[[str, str], float], grid: List[List[float]]) \
        -> Tuple[Seq, Seq, float]:
    """
    computes minimum cost alignment of left and right sequences for the given cost grid and cost function

    :param left: left sequence
    :param right: right sequence
    :param cost_of: left - right symbol substitution/indel cost function
    :param grid: partial edit distance costs grid
    :return: left sequence alignment, right sequence alignment and edit distance
    """
    L_align, R_align = deque(), deque()
    row, col = len(left), len(right)

    # order of conditionals is important as it will determine which alignment (middle[=implemented], left, right)
    # will be preferred when they have the same cost
    while row + col > 0:

        cost = grid[row][col]

        if row > 0 and col > 0 and abs(cost - grid[row - 1][col - 1] - cost_of(left[row - 1], right[col - 1])) < DELTA:

            L_align.appendleft(left[row - 1])
            R_align.appendleft(right[col - 1])
            row -= 1
            col -= 1

        elif row > 0 and abs(cost - grid[row - 1][col] - cost_of(left[row - 1], GAP)) < DELTA:

            L_align.appendleft(left[row - 1])
            R_align.appendleft(GAP)
            row -= 1

        elif col > 0 and abs(cost - grid[row][col - 1] - cost_of(GAP, right[col - 1])) < DELTA:

            L_align.appendleft(GAP)
            R_align.appendleft(right[col - 1])
            col -= 1

        else:
            raise Exception(f'No cell cost match for '
                            f'alignments {L_align} {R_align} of {left} {right} at r=[{row}] c=[{col}]')

    return tuple(L_align), tuple(R_align), grid[-1][-1]


def _grid(left: Sequence[str], right: Sequence[str], cost_of: Fn[[str, str], float]) -> List[List[float]]:
    """
    Builds a 2D-array where each cell contains the cost of the partial paths ending in the
    corresponding positions of the left and right strings. The last cell contains the minimum
    edit distance, aka, alignment cost.

    It follows the convention that the left string is indexed by rows and the right one by columns; also,
    the left-GAP cost is associated with the row above and the GAP-right cost with the column to the left; filling
    of the grid advances by column, top to bottom

    :param left: the left sequence
    :param right: the right sequence
    :param cost_of: the edit operation cost function
    :return: a 2D array filled with the partial alignment costs

    """

    V, H = len(left) + 1, len(right) + 1

    grid = [[0. for c in range(H)] for r in range(V)]

    for r in range(1, V):
        grid[r][0] = grid[r - 1][0] + cost_of(left[r - 1], GAP)  # first column

    for c in range(1, H):

        grid[0][c] = grid[0][c - 1] + cost_of(GAP, right[c - 1])  # first row

        for r in range(1, V):
            grid[r][c] = min(
                grid[r - 1][c - 1] + cost_of(left[r - 1], right[c - 1]),  # diagonal / substitution
                grid[r - 1][c] + cost_of(left[r - 1], GAP),  # vertical / gap in right
                grid[r][c - 1] + cost_of(GAP, right[c - 1]))  # horizontal / gap in left

    return grid


def _join(left: str, right: str) -> str:
    return f'{left}{JOIN}{right}'



from pyaliner.align import align, LEV, COMPACT, CLASSIC, GAP, JOIN, Seq, as_cigar, compact
from pyaliner.compare import compare_in_true_pred_from_file, compare_true_pred_from_file, compare_in_true_pred, \
    compare_true_pred, compare_in_true_pred_from_files, compare_true_pred_from_files
from pyaliner.display import in_terminal, rich_paired_in_true_pred, rich_paired_true_pred, rich_inlined_true_pred

__all__ = ['align',
           'LEV',
           'COMPACT',
           'CLASSIC',
           'GAP',
           'JOIN',
           'Seq',
           'as_cigar',
           'compact',
           'compare_in_true_pred_from_file',
           'compare_in_true_pred_from_files',
           'compare_true_pred_from_file',
           'compare_true_pred_from_files',
           'compare_in_true_pred',
           'compare_true_pred',
           'in_terminal',
           'rich_paired_in_true_pred',
           'rich_paired_true_pred',
           'rich_inlined_true_pred']



"""
API for rich sequence comparisons
"""
from typing import Tuple, Iterable, Sequence

from pypey import pype, px, Fn

from pyaliner import GAP, JOIN
from pyaliner.align import align, COMPACT, CLASSIC, Seq
from pyaliner.display import rich_paired_in_true_pred, rich_paired_true_pred, in_terminal, rich_inlined_true_pred, FILL

compact_align = px(align, kind=COMPACT)

PAIRED = 'paired'
INLINED = 'inlined'


def compare_in_true_pred_from_file(path: str, alignment: str = COMPACT):
    """
    Visually compares triplets of inputs, ground-truths and predictions in the given file.

    :param path: path to file with a collection of three lines, each an input followed by a ground-truth
        followed by a prediction
    :param alignment: the type of alignment to apply to the triplet
    :return: nothing, but displays comparison in terminal
    """
    compare_in_true_pred(pype.file(path).map(str.strip).select(bool).map(str.split), alignment)


def compare_in_true_pred_from_files(in_path: str, true_path: str, pred_path: str, alignment: str = COMPACT):
    """
    Visually compares triplets of inputs, ground-truths and predictions in the given filse.

    :param in_path: path to file with with a line per input sequence
    :param true_path: path to file with with a line per target sequence
    :param pred_path: path to file with with a line per predicted sequence
    :param alignment: the type of alignment to apply to the triplet
    :return: nothing, but displays comparison in terminal
    """
    in_lines, true_lines, pred_lines = \
        (pype.file(path).map(str.strip).select(bool) for path in [in_path, true_path, pred_path])

    compare_in_true_pred(in_lines.interleave(true_lines).interleave(pred_lines, n=2).map(str.split), alignment)


def compare_true_pred_from_file(path: str, alignment: str, view: str = PAIRED):
    """
    Visually compares pairs of ground-truths and predictions in the given file.

    :param path: path to file with an even number of lines, each being a ground-truth or a prediction, in that order
    :param alignment: the type of alignment to apply to the triplet
    :param view: the view to display in the terminal
    :return: nothing, but displays comparison in terminal
    """
    compare_true_pred(pype.file(path).map(str.strip).select(bool).map(str.split), alignment, view)


def compare_true_pred_from_files(true_path: str, pred_path: str, alignment: str, view: str = PAIRED):
    """
    Visually compares pairs of ground-truths and predictions in the given files.

    :param true_path: path to file with with a line per target sequence
    :param pred_path: path to file with with a line per predicted sequence
    :param alignment: the type of alignment to apply to the triplet
    :param view: the view to display in the terminal
    :return: nothing, but displays comparison in terminal
    """
    true_lines, pred_lines = (pype.file(path).map(str.strip).select(bool) for path in [true_path, pred_path])

    compare_true_pred(true_lines.interleave(pred_lines).map(str.split), alignment, view)


def compare_in_true_pred(triplets: Iterable[Iterable[str]], alignment: str = COMPACT):
    """
    Visually compares triplets of inputs, ground-truths and predictions in the collection.

    :param triplets: collection of three strings, each being an input, a ground-truth or a prediction, in that order
    :param alignment: the type of alignment to apply to the triplet
    :return: nothing, but displays comparison in terminal
    """
    align_fn = align if alignment == CLASSIC else compact_align

    alignments = pype(triplets).chunk(size=3).map(lambda in_seq, true, pred: _3way_align(in_seq, true, pred, align_fn))

    in_terminal(rich_paired_in_true_pred(alignments))


def compare_true_pred(pairs: Iterable[Iterable[str]], alignment: str, view: str = PAIRED):
    """
    Visually compares pairs of ground-truths and predictions in the given file.

    :param pairs: collection of even number of strings, each ground-truth followed by a prediction
    :param alignment: the type of alignment to apply to the triplet
    :param view: the view to display in the terminal
    :return: nothing, but displays comparison in terminal
    """

    alignments = pype(pairs).chunk(size=2).map(compact_align if alignment == COMPACT else align)

    in_terminal(rich_inlined_true_pred(alignments) if view == INLINED else rich_paired_true_pred(alignments))


def _3way_align(in_seq: Sequence[str], true: Sequence[str], pred: Sequence[str], align_fn: Fn) -> Tuple[Seq, Seq, Seq]:
    # TODO THIS IS A SLOW HEURISTIC METHOD, NEEDS REPLACING WITH MORE PERFORMANT BETTER THOUGHT-THROUGH ALGORITHM

    true_ali, pred_ali = align_fn(true, pred)

    input_ali, true_ali = align_fn(in_seq, true_ali)
    input_ali, pred_ali = align_fn(input_ali, pred_ali)

    if len(input_ali) != len(true_ali) or len(true_ali) != len(pred_ali):
        true_ali, pred_ali = align_fn(true_ali, pred_ali)

        input_ali, true_ali = align_fn(input_ali, true_ali)
        input_ali, pred_ali = align_fn(input_ali, pred_ali)

    if len(input_ali) != len(true_ali) or len(true_ali) != len(pred_ali):
        true_ali, pred_ali = align_fn(true_ali, pred_ali)

        input_ali, true_ali = align_fn(input_ali, true_ali)
        input_ali, pred_ali = align_fn(input_ali, pred_ali)

    return (pype([input_ali, true_ali, pred_ali])
            .zip(trunc=False, pad=FILL)
            .reject(lambda _in, true, pred: _in == GAP and true == GAP and pred == GAP)
            .map(lambda _in, true, pred: (_in.strip(GAP + JOIN) if len(_in) > 2 else _in, true, pred))
            .unzip()
            .map(tuple)
            .to(tuple))
